@@ -23,6 +23,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <time.h>
 #include <ctype.h>
 #include <assert.h>
 #include <unistd.h>
@@ -40,6 +41,7 @@
 
 #include "cvt-openpgp.h"
 #include "../common/sexp-parse.h"
+#include "../common/openpgpdefs.h"  /* For s2k functions.  */
 
 
 /* The protection mode for encryption.  The supported modes for
@@ -48,9 +50,6 @@
 #define PROT_CIPHER_STRING "aes"
 #define PROT_CIPHER_KEYLEN (128/8)
 
-/* Decode an rfc4880 encoded S2K count.  */
-#define S2K_DECODE_COUNT(_val) ((16ul + ((_val) & 15)) << (((_val) >> 4) + 6))
-
 
 /* A table containing the information needed to create a protected
    private key.  */
@@ -70,6 +69,13 @@ static const struct {
 };
 
 
+/* The number of milliseconds we use in the S2K function and the
+ * calibrated count value.  A count value of zero indicates that the
+ * calibration has not yet been done or needs to be done again.  */
+static unsigned int s2k_calibration_time = AGENT_S2K_CALIBRATION;
+static unsigned long s2k_calibrated_count;
+
+
 /* A helper object for time measurement.  */
 struct calibrate_time_s
 {
@@ -90,25 +96,43 @@ hash_passphrase (const char *passphrase, int hashalgo,
 
 
 
+/*
+ * Determine if we can use clock_gettime with CLOCK_THREAD_CPUTIME_ID,
+ * at compile time.
+ */
+#if defined (CLOCK_THREAD_CPUTIME_ID)
+# if _POSIX_THREAD_CPUTIME > 0
+# define USE_CLOCK_GETTIME 1
+# elif _POSIX_THREAD_CPUTIME == 0
+/*
+ * In this case, we should check sysconf with _POSIX_THREAD_CPUTIME at
+ * run time.  As heuristics, for system with newer GNU C library, we
+ * can assume it is available.
+ */
+#  if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 17
+#  define USE_CLOCK_GETTIME 1
+#  endif
+# endif
+#else
+#undef USE_CLOCK_GETTIME
+#endif
+
 /* Get the process time and store it in DATA.  */
 static void
 calibrate_get_time (struct calibrate_time_s *data)
 {
 #ifdef HAVE_W32_SYSTEM
-# ifdef HAVE_W32CE_SYSTEM
-  GetThreadTimes (GetCurrentThread (),
-                  &data->creation_time, &data->exit_time,
-                  &data->kernel_time, &data->user_time);
-# else
   GetProcessTimes (GetCurrentProcess (),
                    &data->creation_time, &data->exit_time,
                    &data->kernel_time, &data->user_time);
-# endif
-#else
-  struct tms tmp;
+#elif defined (USE_CLOCK_GETTIME)
+  struct timespec tmp;
 
-  times (&tmp);
-  data->ticks = tmp.tms_utime;
+  clock_gettime (CLOCK_THREAD_CPUTIME_ID, &tmp);
+  data->ticks = (clock_t)(((unsigned long long)tmp.tv_sec * 1000000000 +
+                           tmp.tv_nsec) * CLOCKS_PER_SEC / 1000000000);
+#else
+  data->ticks = clock ();
 #endif
 }
 
@@ -135,7 +159,7 @@ calibrate_elapsed_time (struct calibrate_time_s *starttime)
   }
 #else
   return (unsigned long)((((double) (stoptime.ticks - starttime->ticks))
-                          /CLOCKS_PER_SEC)*10000000);
+                          /CLOCKS_PER_SEC)*1000);
 #endif
 }
 
@@ -159,7 +183,7 @@ calibrate_s2k_count_one (unsigned long count)
 
 
 /* Measure the time we need to do the hash operations and deduce an
-   S2K count which requires about 100ms of time.  */
+   S2K count which requires roughly some targeted amount of time.  */
 static unsigned long
 calibrate_s2k_count (void)
 {
@@ -171,11 +195,11 @@ calibrate_s2k_count (void)
       ms = calibrate_s2k_count_one (count);
       if (opt.verbose > 1)
         log_info ("S2K calibration: %lu -> %lums\n", count, ms);
-      if (ms > 100)
+      if (ms > s2k_calibration_time)
         break;
     }
 
-  count = (unsigned long)(((double)count / ms) * 100);
+  count = (unsigned long)(((double)count / ms) * s2k_calibration_time);
   count /= 1024;
   count *= 1024;
   if (count < 65536)
@@ -191,18 +215,30 @@ calibrate_s2k_count (void)
 }
 
 
+/* Set the calibration time.  This may be called early at startup or
+ * at any time.  Thus it should one set variables.  */
+void
+set_s2k_calibration_time (unsigned int milliseconds)
+{
+  if (!milliseconds)
+    milliseconds = AGENT_S2K_CALIBRATION;
+  else if (milliseconds > 60 * 1000)
+    milliseconds = 60 * 1000;  /* Cap at 60 seconds.  */
+  s2k_calibration_time = milliseconds;
+  s2k_calibrated_count = 0;  /* Force re-calibration.  */
+}
+
+
 /* Return the calibrated S2K count.  This is only public for the use
  * of the Assuan getinfo s2k_count_cal command.  */
 unsigned long
 get_calibrated_s2k_count (void)
 {
-  static unsigned long count;
-
-  if (!count)
-    count = calibrate_s2k_count ();
+  if (!s2k_calibrated_count)
+    s2k_calibrated_count = calibrate_s2k_count ();
 
   /* Enforce a lower limit.  */
-  return count < 65536 ? 65536 : count;
+  return s2k_calibrated_count < 65536 ? 65536 : s2k_calibrated_count;
 }
 
 
@@ -343,12 +379,11 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
                const char *passphrase,
                const char *timestamp_exp, size_t timestamp_exp_len,
                unsigned char **result, size_t *resultlen,
-	       unsigned long s2k_count, int use_ocb)
+	       unsigned long s2k_count)
 {
   gcry_cipher_hd_t hd;
   const char *modestr;
-  unsigned char hashvalue[20];
-  int blklen, enclen, outlen;
+  int enclen, outlen;
   unsigned char *iv = NULL;
   unsigned int ivsize;  /* Size of the buffer allocated for IV.  */
   const unsigned char *s2ksalt; /* Points into IV.  */
@@ -362,44 +397,26 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
   *resultlen = 0;
   *result = NULL;
 
-  modestr = (use_ocb? "openpgp-s2k3-ocb-aes"
-             /*   */: "openpgp-s2k3-sha1-" PROT_CIPHER_STRING "-cbc");
+  modestr = "openpgp-s2k3-ocb-aes";
 
   rc = gcry_cipher_open (&hd, PROT_CIPHER,
-                         use_ocb? GCRY_CIPHER_MODE_OCB :
-                         GCRY_CIPHER_MODE_CBC,
+                         GCRY_CIPHER_MODE_OCB,
                          GCRY_CIPHER_SECURE);
   if (rc)
     return rc;
 
   /* We need to work on a copy of the data because this makes it
    * easier to add the trailer and the padding and more important we
-   * have to prefix the text with 2 parenthesis.  In CBC mode we
-   * have to allocate enough space for:
-   *
-   *   ((<parameter_list>)(4:hash4:sha120:<hashvalue>)) + padding
-   *
-   * we always append a full block of random bytes as padding but
-   * encrypt only what is needed for a full blocksize.  In OCB mode we
+   * have to prefix the text with 2 parenthesis.  Due to OCB mode we
    * have to allocate enough space for just:
    *
    *   ((<parameter_list>))
    */
-  blklen = gcry_cipher_get_algo_blklen (PROT_CIPHER);
-  if (use_ocb)
-    {
-      /*       ((            )) */
-      outlen = 2 + protlen + 2 ;
-      enclen = outlen + 16 /* taglen */;
-      outbuf = gcry_malloc_secure (enclen);
-    }
-  else
-    {
-      /*       ((            )( 4:hash 4:sha1 20:<hash> ))  <padding>  */
-      outlen = 2 + protlen + 2 + 6   + 6    + 23      + 2 + blklen;
-      enclen = outlen/blklen * blklen;
-      outbuf = gcry_malloc_secure (outlen);
-    }
+
+  /*       ((            )) */
+  outlen = 2 + protlen + 2 ;
+  enclen = outlen + 16 /* taglen */;
+  outbuf = gcry_malloc_secure (enclen);
   if (!outbuf)
     {
       rc = out_of_core ();
@@ -409,10 +426,10 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
   /* Allocate a buffer for the nonce and the salt.  */
   if (!rc)
     {
-      /* Allocate random bytes to be used as IV, padding and s2k salt
-       * or in OCB mode for a nonce and the s2k salt.  The IV/nonce is
-       * set later because for OCB we need to set the key first.  */
-      ivsize = (use_ocb? 12 : (blklen*2)) + 8;
+      /* Allocate random bytes to be used as nonce and s2k salt.  The
+       * nonce is set later because for OCB we need to set the key
+       * first.  */
+      ivsize = 12 + 8;
       iv = xtrymalloc (ivsize);
       if (!iv)
         rc = gpg_error_from_syserror ();
@@ -448,40 +465,17 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
     goto leave;
 
   /* Set the IV/nonce.  */
-  rc = gcry_cipher_setiv (hd, iv, use_ocb? 12 : blklen);
+  rc = gcry_cipher_setiv (hd, iv, 12);
   if (rc)
     goto leave;
 
-  if (use_ocb)
-    {
-      /* In OCB Mode we use only the public key parameters as AAD.  */
-      rc = gcry_cipher_authenticate (hd, hashbegin, protbegin - hashbegin);
-      if (!rc)
-        rc = gcry_cipher_authenticate (hd, timestamp_exp, timestamp_exp_len);
-      if (!rc)
-        rc = gcry_cipher_authenticate
-          (hd, protbegin+protlen, hashlen - (protbegin+protlen - hashbegin));
-    }
-  else
-    {
-      /* Hash the entire expression for CBC mode.  Because
-       * TIMESTAMP_EXP won't get protected, we can't simply hash a
-       * continuous buffer but need to call md_write several times.  */
-      gcry_md_hd_t md;
-
-      rc = gcry_md_open (&md, GCRY_MD_SHA1, 0 );
-      if (!rc)
-        {
-          gcry_md_write (md, hashbegin, protbegin - hashbegin);
-          gcry_md_write (md, protbegin, protlen);
-          gcry_md_write (md, timestamp_exp, timestamp_exp_len);
-          gcry_md_write (md, protbegin+protlen,
-                         hashlen - (protbegin+protlen - hashbegin));
-          memcpy (hashvalue, gcry_md_read (md, GCRY_MD_SHA1), 20);
-          gcry_md_close (md);
-        }
-    }
-
+  /* In OCB Mode we use only the public key parameters as AAD.  */
+  rc = gcry_cipher_authenticate (hd, hashbegin, protbegin - hashbegin);
+  if (!rc)
+    rc = gcry_cipher_authenticate (hd, timestamp_exp, timestamp_exp_len);
+  if (!rc)
+    rc = gcry_cipher_authenticate
+      (hd, protbegin+protlen, hashlen - (protbegin+protlen - hashbegin));
 
   /* Encrypt.  */
   if (!rc)
@@ -491,36 +485,15 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
       *p++ = '(';
       memcpy (p, protbegin, protlen);
       p += protlen;
-      if (use_ocb)
+      *p++ = ')';
+      *p++ = ')';
+      log_assert ( p - outbuf == outlen);
+      gcry_cipher_final (hd);
+      rc = gcry_cipher_encrypt (hd, outbuf, outlen, NULL, 0);
+      if (!rc)
         {
-          *p++ = ')';
-          *p++ = ')';
-        }
-      else
-        {
-          memcpy (p, ")(4:hash4:sha120:", 17);
-          p += 17;
-          memcpy (p, hashvalue, 20);
-          p += 20;
-          *p++ = ')';
-          *p++ = ')';
-          memcpy (p, iv+blklen, blklen); /* Add padding.  */
-          p += blklen;
-        }
-      assert ( p - outbuf == outlen);
-      if (use_ocb)
-        {
-          gcry_cipher_final (hd);
-          rc = gcry_cipher_encrypt (hd, outbuf, outlen, NULL, 0);
-          if (!rc)
-            {
-              log_assert (outlen + 16 == enclen);
-              rc = gcry_cipher_gettag (hd, outbuf + outlen, 16);
-            }
-        }
-      else
-        {
-          rc = gcry_cipher_encrypt (hd, outbuf, enclen, NULL, 0);
+          log_assert (outlen + 16 == enclen);
+          rc = gcry_cipher_gettag (hd, outbuf + outlen, 16);
         }
     }
 
@@ -548,7 +521,7 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
        (int)strlen (modestr), modestr,
        &saltpos,
        (unsigned int)strlen (countbuf), countbuf,
-       use_ocb? 12 : blklen, &ivpos, use_ocb? 12 : blklen, "",
+       12, &ivpos, 12, "",
        enclen, &encpos, enclen, "");
     if (!p)
       {
@@ -562,7 +535,7 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
   *resultlen = strlen (p);
   *result = (unsigned char*)p;
   memcpy (p+saltpos, s2ksalt, 8);
-  memcpy (p+ivpos, iv, use_ocb? 12 : blklen);
+  memcpy (p+ivpos, iv, 12);
   memcpy (p+encpos, outbuf, enclen);
   xfree (iv);
   xfree (outbuf);
@@ -578,13 +551,11 @@ do_encryption (const unsigned char *hashbegin, size_t hashlen,
 
 
 /* Protect the key encoded in canonical format in PLAINKEY.  We assume
-   a valid S-Exp here.  With USE_UCB set to -1 the default scheme is
-   used (ie. either CBC or OCB), set to 0 the old CBC mode is used,
-   and set to 1 OCB is used. */
+ * a valid S-Exp here.  */
 int
 agent_protect (const unsigned char *plainkey, const char *passphrase,
                unsigned char **result, size_t *resultlen,
-	       unsigned long s2k_count, int use_ocb)
+	       unsigned long s2k_count)
 {
   int rc;
   const char *parmlist;
@@ -601,9 +572,6 @@ agent_protect (const unsigned char *plainkey, const char *passphrase,
   unsigned char *p;
   int have_curve = 0;
 
-  if (use_ocb == -1)
-    use_ocb = opt.enable_extended_key_format;
-
   /* Create an S-expression with the protected-at timestamp.  */
   memcpy (timestamp_exp, "(12:protected-at15:", 19);
   gnupg_get_isotime (timestamp_exp+19);
@@ -697,17 +665,17 @@ agent_protect (const unsigned char *plainkey, const char *passphrase,
   hash_end = s;
   s++;
   /* Skip to the end of the S-expression.  */
-  assert (depth == 1);
+  log_assert (depth == 1);
   rc = sskip (&s, &depth);
   if (rc)
     return rc;
-  assert (!depth);
+  log_assert (!depth);
   real_end = s-1;
 
   rc = do_encryption (hash_begin, hash_end - hash_begin + 1,
                       prot_begin, prot_end - prot_begin + 1,
                       passphrase, timestamp_exp, sizeof (timestamp_exp),
-                      &protected, &protectedlen, s2k_count, use_ocb);
+                      &protected, &protectedlen, s2k_count);
   if (rc)
     return rc;
 
@@ -739,7 +707,7 @@ agent_protect (const unsigned char *plainkey, const char *passphrase,
 
   memcpy (p, prot_end+1, real_end - prot_end);
   p += real_end - prot_end;
-  assert ( p - *result == *resultlen);
+  log_assert ( p - *result == *resultlen);
   xfree (protected);
 
   return 0;
@@ -928,7 +896,10 @@ merge_lists (const unsigned char *protectedkey,
   /* Copy the cleartext.  */
   s = cleartext;
   if (*s != '(' && s[1] != '(')
-    return gpg_error (GPG_ERR_BUG);  /*we already checked this */
+    {
+      xfree (newlist);
+      return gpg_error (GPG_ERR_BUG);  /*we already checked this */
+    }
   s += 2;
   startpos = s;
   while ( *s == '(' )
@@ -978,7 +949,7 @@ merge_lists (const unsigned char *protectedkey,
 
   /* Skip over the protected list element in the original list.  */
   s = protectedkey + replacepos;
-  assert (*s == '(');
+  log_assert (*s == '(');
   s++;
   i = 1;
   rc = sskip (&s, &i);
@@ -1005,7 +976,7 @@ merge_lists (const unsigned char *protectedkey,
   rc = sskip (&s, &i);
   if (rc)
     goto failure;
-  assert (s[-1] == ')');
+  log_assert (s[-1] == ')');
   endpos = s; /* one behind the end of the list */
 
   /* Append the rest. */
@@ -1105,7 +1076,7 @@ agent_unprotect (ctrl_t ctrl,
   if (!protect_info[infidx].algo)
     return gpg_error (GPG_ERR_UNSUPPORTED_ALGORITHM);
 
-  /* See wether we have a protected-at timestamp.  */
+  /* See whether we have a protected-at timestamp.  */
   protect_list = s;  /* Save for later.  */
   if (protected_at)
     {
@@ -1490,14 +1461,15 @@ make_shadow_info (const char *serialno, const char *idstring)
 
 
 /* Create a shadow key from a public key.  We use the shadow protocol
-  "ti-v1" and insert the S-expressionn SHADOW_INFO.  The resulting
+  "t1-v1" and insert the S-expressionn SHADOW_INFO.  The resulting
   S-expression is returned in an allocated buffer RESULT will point
   to. The input parameters are expected to be valid canonicalized
   S-expressions */
 int
-agent_shadow_key (const unsigned char *pubkey,
-                  const unsigned char *shadow_info,
-                  unsigned char **result)
+agent_shadow_key_type (const unsigned char *pubkey,
+                       const unsigned char *shadow_info,
+                       const unsigned char *type,
+                       unsigned char **result)
 {
   const unsigned char *s;
   const unsigned char *point;
@@ -1550,10 +1522,10 @@ agent_shadow_key (const unsigned char *pubkey,
   point = s; /* insert right before the point */
   depth--;
   s++;
-  assert (depth == 1);
+  log_assert (depth == 1);
 
   /* Calculate required length by taking in account: the "shadowed-"
-     prefix, the "shadowed", "t1-v1" as well as some parenthesis */
+     prefix, the "shadowed", shadow type as well as some parenthesis */
   n = 12 + pubkey_len + 1 + 3+8 + 2+5 + shadow_info_len + 1;
   *result = xtrymalloc (n);
   p = (char*)*result;
@@ -1563,7 +1535,7 @@ agent_shadow_key (const unsigned char *pubkey,
   /* (10:public-key ...)*/
   memcpy (p, pubkey+14, point - (pubkey+14));
   p += point - (pubkey+14);
-  p = stpcpy (p, "(8:shadowed5:t1-v1");
+  p += sprintf (p, "(8:shadowed%d:%s", (int)strlen(type), type);
   memcpy (p, shadow_info, shadow_info_len);
   p += shadow_info_len;
   *p++ = ')';
@@ -1573,16 +1545,26 @@ agent_shadow_key (const unsigned char *pubkey,
   return 0;
 }
 
-/* Parse a canonical encoded shadowed key and return a pointer to the
-   inner list with the shadow_info */
-gpg_error_t
-agent_get_shadow_info (const unsigned char *shadowkey,
-                       unsigned char const **shadow_info)
+int
+agent_shadow_key (const unsigned char *pubkey,
+                  const unsigned char *shadow_info,
+                  unsigned char **result)
 {
-  const unsigned char *s;
-  size_t n;
+  return agent_shadow_key_type (pubkey, shadow_info, "t1-v1", result);
+}
+
+/* Parse a canonical encoded shadowed key and return a pointer to the
+   inner list with the shadow_info and the shadow type */
+gpg_error_t
+agent_get_shadow_info_type (const unsigned char *shadowkey,
+                            unsigned char const **shadow_info,
+                            unsigned char **shadow_type)
+{
+  const unsigned char *s, *saved_s;
+  size_t n, saved_n;
   int depth = 0;
 
+  (void)depth;
   s = shadowkey;
   if (*s != '(')
     return gpg_error (GPG_ERR_INV_SEXP);
@@ -1629,24 +1611,80 @@ agent_get_shadow_info (const unsigned char *shadowkey,
   n = snext (&s);
   if (!n)
     return gpg_error (GPG_ERR_INV_SEXP);
-  if (smatch (&s, n, "t1-v1"))
+  saved_s = s;
+  saved_n = n;
+  if (smatch (&s, n, "t1-v1") || smatch(&s, n, "tpm2-v1"))
     {
       if (*s != '(')
         return gpg_error (GPG_ERR_INV_SEXP);
-      *shadow_info = s;
+      if (shadow_info)
+        *shadow_info = s;
     }
   else
     return gpg_error (GPG_ERR_UNSUPPORTED_PROTOCOL);
+  s = saved_s;
+  n = saved_n;
+
+  if (shadow_type)
+    {
+      char *buf = xtrymalloc(n+1);
+      if (!buf)
+        return gpg_error_from_syserror ();
+      memcpy (buf, s, n);
+      buf[n] = '\0';
+      *shadow_type = buf;
+    }
+
   return 0;
 }
 
 
+gpg_error_t
+agent_get_shadow_info (const unsigned char *shadowkey,
+                       unsigned char const **shadow_info)
+{
+  return agent_get_shadow_info_type (shadowkey, shadow_info, NULL);
+}
+
+
+int
+agent_is_tpm2_key (gcry_sexp_t s_skey)
+{
+  unsigned char *buf;
+  unsigned char *type;
+  size_t len;
+  gpg_error_t err;
+
+  err = make_canon_sexp (s_skey, &buf, &len);
+  if (err)
+    return 0;
+
+  err = agent_get_shadow_info_type (buf, NULL, &type);
+  xfree (buf);
+  if (err)
+    return 0;
+
+  err = strcmp (type, "tpm2-v1") == 0;
+  xfree (type);
+  return err;
+}
+
+
+gpg_error_t
+agent_get_shadow_type (const unsigned char *shadowkey,
+                       unsigned char **shadow_type)
+{
+  return agent_get_shadow_info_type (shadowkey, NULL, shadow_type);
+}
+
+
 /* Parse the canonical encoded SHADOW_INFO S-expression.  On success
    the hex encoded serial number is returned as a malloced strings at
    R_HEXSN and the Id string as a malloced string at R_IDSTR.  On
    error an error code is returned and NULL is stored at the result
    parameters addresses.  If the serial number or the ID string is not
-   required, NULL may be passed for them.  */
+   required, NULL may be passed for them.  Note that R_PINLEN is
+   currently not used by any caller.  */
 gpg_error_t
 parse_shadow_info (const unsigned char *shadow_info,
                    char **r_hexsn, char **r_idstr, int *r_pinlen)
