@@ -80,7 +80,7 @@ static void writeout_rem (audit_ctx_t ctx,
 
 
 /* Add NAME to the list of help tags.  NAME needs to be a const string
-   an this function merly stores this pointer.  */
+   an this function merely stores this pointer.  */
 static void
 add_helptag (audit_ctx_t ctx, const char *name)
 {
@@ -716,7 +716,7 @@ list_cert (audit_ctx_t ctx, ksba_cert_t cert, int with_subj)
 
 
 /* List the chain of certificates from STARTITEM up to STOPEVENT.  The
-   certifcates are written out as comments.  */
+   certificates are written out as comments.  */
 static void
 list_certchain (audit_ctx_t ctx, log_item_t startitem, audit_event_t stopevent)
 {
@@ -1105,9 +1105,11 @@ proc_type_verify (audit_ctx_t ctx)
           switch (gpg_err_code (item->err))
             {
             case 0:                    ok = "good"; break;
+            case GPG_ERR_TRUE:         ok = "n/a";  break;
             case GPG_ERR_CERT_REVOKED: ok = "bad"; break;
             case GPG_ERR_NOT_ENABLED:  ok = "disabled"; break;
             case GPG_ERR_NO_CRL_KNOWN:
+            case GPG_ERR_INV_CRL_OBJ:
               ok = _("no CRL found for certificate");
               break;
             case GPG_ERR_CRL_TOO_OLD:
