@@ -16,6 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
@@ -32,15 +33,14 @@
 #include <assuan.h>
 
 #include "../common/logging.h"
-#include "../common/argparse.h"
 #include "../common/stringhelp.h"
 #include "../common/mischelp.h"
 #include "../common/strlist.h"
 #include "../common/asshelp.h"
 
-#include "i18n.h"
-#include "util.h"
-#include "init.h"
+#include "../common/i18n.h"
+#include "../common/util.h"
+#include "../common/init.h"
 
 
 /* Constants for the options.  */
@@ -65,7 +65,7 @@ enum
 
 
 /* The list of options as used by the argparse.c code.  */
-static ARGPARSE_OPTS opts[] = {
+static gpgrt_opt_t opts[] = {
   { oVerbose,  "verbose",   0, N_("verbose") },
   { oQuiet,    "quiet",     0, N_("be somewhat more quiet") },
   { oOCSP,     "ocsp",      0, N_("use OCSP instead of CRLs") },
@@ -80,7 +80,7 @@ static ARGPARSE_OPTS opts[] = {
   { oPEM,      "pem",       0, N_("expect certificates in PEM format")},
   { oForceDefaultResponder, "force-default-responder", 0,
     N_("force the use of the default OCSP responder")},
-  { 0, NULL, 0, NULL }
+  ARGPARSE_end ()
 };
 
 
@@ -158,9 +158,11 @@ my_strusage (int level)
 
   switch(level)
     {
+    case  9: p = "GPL-3.0-or-later"; break;
     case 11: p = "dirmngr-client (@GNUPG@)";
       break;
     case 13: p = VERSION; break;
+    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
     case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
     case 49: p = PACKAGE_BUGREPORT; break;
@@ -186,7 +188,7 @@ my_strusage (int level)
 int
 main (int argc, char **argv )
 {
-  ARGPARSE_ARGS pargs;
+  gpgrt_argparse_t pargs;
   assuan_context_t ctx;
   gpg_error_t err;
   unsigned char *certbuf;
@@ -199,9 +201,12 @@ main (int argc, char **argv )
   int cmd_squid_mode = 0;
 
   early_system_init ();
-  set_strusage (my_strusage);
+  gpgrt_set_strusage (my_strusage);
   log_set_prefix ("dirmngr-client",
                   GPGRT_LOG_WITH_PREFIX);
+  /* Register our string mapper with gpgrt.  Usually done in
+   * init_common_subsystems, but we don't need that here.  */
+  gpgrt_set_fixed_string_mapper (map_static_macro_string);
 
   /* For W32 we need to initialize the socket subsystem.  Because we
      don't use Pth we need to do this explicit. */
@@ -223,8 +228,8 @@ main (int argc, char **argv )
   /* Parse the command line.  */
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags= 1;  /* Do not remove the args. */
-  while (arg_parse (&pargs, opts) )
+  pargs.flags= ARGPARSE_FLAG_KEEP;
+  while (gpgrt_argparse (NULL, &pargs, opts))
     {
       switch (pargs.r_opt)
         {
@@ -247,9 +252,11 @@ main (int argc, char **argv )
           break;
         case oForceDefaultResponder: opt.force_default_responder = 1; break;
 
-        default : pargs.err = 2; break;
+        default : pargs.err = ARGPARSE_PRINT_ERROR; break;
 	}
     }
+  gpgrt_argparse (NULL, &pargs, NULL);
+
   if (log_get_errorcount (0))
     exit (2);
 
@@ -258,14 +265,14 @@ main (int argc, char **argv )
   else if (cmd_lookup || cmd_loadcrl)
     {
       if (!argc)
-        usage (1);
+        gpgrt_usage (1);
       err = 0;
     }
   else if (cmd_squid_mode)
     {
       err = 0;
       if (argc)
-        usage (1);
+        gpgrt_usage (1);
     }
   else if (!argc)
     {
@@ -284,7 +291,7 @@ main (int argc, char **argv )
   else
     {
       err = 0;
-      usage (1);
+      gpgrt_usage (1);
     }
 
   if (log_get_errorcount (0))
@@ -452,9 +459,10 @@ data_cb (void *opaque, const void *buffer, size_t length)
    returned in an alloced buffer whose address will be returned in
    RBUF and its length in RBUFLEN.  */
 static gpg_error_t
-read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
+read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen,
+                      int no_errmsg)
 {
-  FILE *fp;
+  estream_t fp;
   int c;
   int pos;
   int value;
@@ -468,16 +476,16 @@ read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
 
   init_asctobin ();
 
-  fp = fname? fopen (fname, "r") : stdin;
+  fp = fname? es_fopen (fname, "r") : es_stdin;
   if (!fp)
-    return gpg_error_from_errno (errno);
+    return gpg_error_from_syserror ();
 
   pos = 0;
   value = 0;
   bufsize = 8192;
   buf = xmalloc (bufsize);
   buflen = 0;
-  while ((c=getc (fp)) != EOF)
+  while ((c=es_getc (fp)) != EOF)
     {
       int escaped_c = 0;
 
@@ -486,10 +494,10 @@ read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
           if (c == '%')
             {
               char tmp[2];
-              if ((c = getc(fp)) == EOF)
+              if ((c = es_getc(fp)) == EOF)
                 break;
               tmp[0] = c;
-              if ((c = getc(fp)) == EOF)
+              if ((c = es_getc(fp)) == EOF)
                 break;
               tmp[1] = c;
               if (!hexdigitp (tmp) || !hexdigitp (tmp+1))
@@ -497,7 +505,7 @@ read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
                   log_error ("invalid percent escape sequence\n");
                   state = s_idle; /* Force an error. */
                   /* Skip to end of line.  */
-                  while ( (c=getc (fp)) != EOF && c != '\n')
+                  while ( (c=es_getc (fp)) != EOF && c != '\n')
                     ;
                   goto ready;
                 }
@@ -517,7 +525,7 @@ read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
             }
           break;
         case s_init:
-          state = s_lfseen;
+          state = s_lfseen; /* fall through */
         case s_lfseen:
           if (c != "-----BEGIN "[pos])
             state = s_idle;
@@ -586,7 +594,7 @@ read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
     }
  ready:
   if (fname)
-    fclose (fp);
+    es_fclose (fp);
 
   if (state == s_init && c == EOF)
     {
@@ -595,7 +603,8 @@ read_pem_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
     }
   else if (state != s_waitend)
     {
-      log_error ("no certificate or invalid encoded\n");
+      if (!no_errmsg)
+        log_error ("no certificate or invalid encoded\n");
       xfree (buf);
       return gpg_error (GPG_ERR_INV_ARMOR);
     }
@@ -613,25 +622,27 @@ static gpg_error_t
 read_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
 {
   gpg_error_t err;
-  FILE *fp;
+  estream_t fp;
   unsigned char *buf;
   size_t nread, bufsize, buflen;
 
   if (opt.pem)
-    return read_pem_certificate (fname, rbuf, rbuflen);
+    return read_pem_certificate (fname, rbuf, rbuflen, 0);
   else if (fname)
     {
       /* A filename has been given.  Let's just assume it is in PEM
          format and decode it, and fall back to interpreting it as
          binary certificate if that fails.  */
-      err = read_pem_certificate (fname, rbuf, rbuflen);
+      err = read_pem_certificate (fname, rbuf, rbuflen, 1);
       if (! err)
         return 0;
+      /* Clear the error count to try as binary certificate.  */
+      log_get_errorcount (1);
     }
 
-  fp = fname? fopen (fname, "rb") : stdin;
+  fp = fname? es_fopen (fname, "rb") : es_stdin;
   if (!fp)
-    return gpg_error_from_errno (errno);
+    return gpg_error_from_syserror ();
 
   buf = NULL;
   bufsize = buflen = 0;
@@ -644,13 +655,13 @@ read_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
       else
         buf = xrealloc (buf, bufsize);
 
-      nread = fread (buf+buflen, 1, NCHUNK, fp);
-      if (nread < NCHUNK && ferror (fp))
+      nread = es_fread (buf+buflen, 1, NCHUNK, fp);
+      if (nread < NCHUNK && es_ferror (fp))
         {
-          err = gpg_error_from_errno (errno);
+          err = gpg_error_from_syserror ();
           xfree (buf);
           if (fname)
-            fclose (fp);
+            es_fclose (fp);
           return err;
         }
       buflen += nread;
@@ -658,7 +669,7 @@ read_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
   while (nread == NCHUNK);
 #undef NCHUNK
   if (fname)
-    fclose (fp);
+    es_fclose (fp);
   *rbuf = buf;
   *rbuflen = buflen;
   return 0;
@@ -896,7 +907,7 @@ squid_loop_body (assuan_context_t ctx)
   unsigned char *certbuf;
   size_t certbuflen = 0;
 
-  err = read_pem_certificate (NULL, &certbuf, &certbuflen);
+  err = read_pem_certificate (NULL, &certbuf, &certbuflen, 0);
   if (gpg_err_code (err) == GPG_ERR_EOF)
     return err;
   if (err)
