@@ -116,10 +116,15 @@ read_response (estream_t fp, unsigned char **r_buffer, size_t *r_buflen)
 
 /* Construct an OCSP request, send it to the configured OCSP responder
    and parse the response. On success the OCSP context may be used to
-   further process the response. */
+   further process the response.  The signature value and the
+   production date are returned at R_SIGVAL and R_PRODUCED_AT; they
+   may be NULL or an empty string if not available.  A new hash
+   context is returned at R_MD.  */
 static gpg_error_t
-do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
-                 const char *url, ksba_cert_t cert, ksba_cert_t issuer_cert)
+do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp,
+                 const char *url, ksba_cert_t cert, ksba_cert_t issuer_cert,
+                 ksba_sexp_t *r_sigval, ksba_isotime_t r_produced_at,
+                 gcry_md_hd_t *r_md)
 {
   gpg_error_t err;
   unsigned char *request, *response;
@@ -132,12 +137,19 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
 
   (void)ctrl;
 
+  *r_sigval = NULL;
+  *r_produced_at = 0;
+  *r_md = NULL;
+
   if (dirmngr_use_tor ())
     {
       /* For now we do not allow OCSP via Tor due to possible privacy
          concerns.  Needs further research.  */
-      log_error (_("OCSP request not possible due to Tor mode\n"));
-      return gpg_error (GPG_ERR_NOT_SUPPORTED);
+      const char *msg = _("OCSP request not possible due to Tor mode");
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      log_error ("%s", msg);
+      dirmngr_status_printf (ctrl, "NOTE", "no_ocsp_due_to_tor %u %s", err,msg);
+      return err;
     }
 
   if (opt.disable_http)
@@ -172,7 +184,7 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
     }
 
  once_more:
-  err = http_open (&http, HTTP_REQ_POST, url, NULL, NULL,
+  err = http_open (ctrl, &http, HTTP_REQ_POST, url, NULL, NULL,
                    ((opt.honor_http_proxy? HTTP_FLAG_TRY_PROXY:0)
                     | (dirmngr_use_tor ()? HTTP_FLAG_FORCE_TOR:0)
                     | (opt.disable_ipv4? HTTP_FLAG_IGNORE_IPv4 : 0)
@@ -215,7 +227,7 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
             case 301:
             case 302:
               {
-                const char *s = http_get_header (http, "Location");
+                const char *s = http_get_header (http, "Location", 0);
 
                 log_info (_("URL '%s' redirected to '%s' (%u)\n"),
                           url, s?s:"[none]", http_get_status_code (http));
@@ -238,6 +250,10 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
               }
               break;
 
+            case 413:  /* Payload too large */
+              err = gpg_error (GPG_ERR_TOO_LARGE);
+              break;
+
             default:
               log_error (_("error accessing '%s': http status %u\n"),
                          url, http_get_status_code (http));
@@ -259,6 +275,7 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
       xfree (free_this);
       return err;
     }
+  /* log_printhex (response, responselen, "ocsp response"); */
 
   err = ksba_ocsp_parse_response (ocsp, response, responselen,
                                   &response_status);
@@ -286,11 +303,34 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
     }
   if (response_status == KSBA_OCSP_RSPSTATUS_SUCCESS)
     {
+      int hash_algo;
+
       if (opt.verbose)
         log_info (_("OCSP responder at '%s' status: %s\n"), url, t);
 
+      /* Get the signature value now because we can call this function
+       * only once.  */
+      *r_sigval = ksba_ocsp_get_sig_val (ocsp, r_produced_at);
+
+      hash_algo = hash_algo_from_sigval (*r_sigval);
+      if (!hash_algo)
+        {
+          if (opt.verbose)
+            log_info ("ocsp: using SHA-256 as fallback hash algo.\n");
+          hash_algo = GCRY_MD_SHA256;
+        }
+      err = gcry_md_open (r_md, hash_algo, 0);
+      if (err)
+        {
+          log_error (_("failed to establish a hashing context for OCSP: %s\n"),
+                     gpg_strerror (err));
+          goto leave;
+        }
+      if (DBG_HASHING)
+        gcry_md_debug (*r_md, "ocsp");
+
       err = ksba_ocsp_hash_response (ocsp, response, responselen,
-                                     HASH_FNC, md);
+                                     HASH_FNC, *r_md);
       if (err)
         log_error (_("hashing the OCSP response for '%s' failed: %s\n"),
                    url, gpg_strerror (err));
@@ -301,8 +341,17 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
       err = gpg_error (GPG_ERR_GENERAL);
     }
 
+ leave:
   xfree (response);
   xfree (free_this);
+  if (err)
+    {
+      xfree (*r_sigval);
+      *r_sigval = NULL;
+      *r_produced_at = 0;
+      gcry_md_close (*r_md);
+      *r_md = NULL;
+    }
   return err;
 }
 
@@ -343,7 +392,7 @@ validate_responder_cert (ctrl_t ctrl, ksba_cert_t cert,
 
          Note, that in theory we could simply ask the client via an
          inquire to validate a certificate but this might involve
-         calling DirMngr again recursivly - we can't do that as of now
+         calling DirMngr again recursively - we can't do that as of now
          (neither DirMngr nor gpgsm have the ability for concurrent
          access to DirMngr.   */
 
@@ -360,38 +409,121 @@ validate_responder_cert (ctrl_t ctrl, ksba_cert_t cert,
 }
 
 
-/* Helper for check_signature. */
-static int
+/* Helper for check_signature.  MD is the finalized hash context.  */
+static gpg_error_t
 check_signature_core (ctrl_t ctrl, ksba_cert_t cert, gcry_sexp_t s_sig,
-                      gcry_sexp_t s_hash, fingerprint_list_t signer_fpr_list)
+                      gcry_md_hd_t md, fingerprint_list_t signer_fpr_list)
 {
   gpg_error_t err;
-  ksba_sexp_t pubkey;
   gcry_sexp_t s_pkey = NULL;
+  gcry_sexp_t s_hash = NULL;
+  const char *s;
+  int mdalgo, mdlen;
 
-  pubkey = ksba_cert_get_public_key (cert);
-  if (!pubkey)
-    err = gpg_error (GPG_ERR_INV_OBJ);
-  else
-    err = canon_sexp_to_gcry (pubkey, &s_pkey);
-  xfree (pubkey);
-  if (!err)
-    err = gcry_pk_verify (s_sig, s_hash, s_pkey);
-  if (!err)
-    err = validate_responder_cert (ctrl, cert, signer_fpr_list);
-  if (!err)
+  /* Get the public key as a gcrypt s-expression.  */
+  {
+    ksba_sexp_t pk = ksba_cert_get_public_key (cert);
+    if (!pk)
+      err = gpg_error (GPG_ERR_INV_OBJ);
+    else
+      {
+        err = canon_sexp_to_gcry (pk, &s_pkey);
+        xfree (pk);
+      }
+    if (err)
+      goto leave;
+  }
+
+  mdalgo = gcry_md_get_algo (md);
+  mdlen  = gcry_md_get_algo_dlen (mdalgo);
+
+  if (pk_algo_from_sexp (s_pkey) == GCRY_PK_ECC)
     {
-      gcry_sexp_release (s_pkey);
-      return 0; /* Successfully verified the signature. */
+      unsigned int qbits0, qbits;
+
+      qbits0 = gcry_pk_get_nbits (s_pkey);
+      qbits = qbits0 == 521? 512 : qbits0;
+
+      if ((qbits%8))
+	{
+	  log_error ("ECDSA requires the hash length to be a"
+                     " multiple of 8 bits\n");
+	  err = gpg_error (GPG_ERR_INTERNAL);
+          goto leave;
+	}
+
+      /* Don't allow any Q smaller than 160 bits.  */
+      if (qbits < 160)
+	{
+	  log_error (_("%s key uses an unsafe (%u bit) hash\n"),
+                     "ECDSA", qbits0);
+	  err = gpg_error (GPG_ERR_INTERNAL);
+          goto leave;
+	}
+
+      /* Check if we're too short.  */
+      if (mdlen < qbits/8)
+        {
+	  log_error (_("a %u bit hash is not valid for a %u bit %s key\n"),
+                     (unsigned int)mdlen*8,
+                     qbits0,
+                     "ECDSA");
+          if (mdlen < 20)
+            {
+              err = gpg_error (GPG_ERR_INTERNAL);
+              goto leave;
+            }
+        }
+
+      /* Truncate.  */
+      if (mdlen > qbits/8)
+        mdlen = qbits/8;
+
+      err = gcry_sexp_build (&s_hash, NULL, "(data(flags raw)(value %b))",
+                             (int)mdlen, gcry_md_read (md, mdalgo));
+    }
+  else if (mdalgo && (s = gcry_md_algo_name (mdalgo)) && strlen (s) < 16)
+    {
+      /* Assume RSA */
+      char hashalgostr[16+1];
+      int i;
+
+      for (i=0; s[i]; i++)
+        hashalgostr[i] = ascii_tolower (s[i]);
+      hashalgostr[i] = 0;
+      err = gcry_sexp_build (&s_hash, NULL, "(data(flags pkcs1)(hash %s %b))",
+                             hashalgostr,
+                             (int)mdlen,
+                             gcry_md_read (md, mdalgo));
+    }
+  else
+    err = gpg_error (GPG_ERR_DIGEST_ALGO);
+  if (err)
+    {
+      log_error (_("creating S-expression failed: %s\n"), gcry_strerror (err));
+      goto leave;
     }
 
-  /* We simply ignore all errors. */
+  if (DBG_CRYPTO)
+    {
+      gcry_log_debugsxp ("sig ", s_sig);
+      gcry_log_debugsxp ("hash", s_hash);
+    }
+
+  err = gcry_pk_verify (s_sig, s_hash, s_pkey);
+  if (err)
+    goto leave;
+
+  err = validate_responder_cert (ctrl, cert, signer_fpr_list);
+
+ leave:
+  gcry_sexp_release (s_hash);
   gcry_sexp_release (s_pkey);
-  return -1;
+  return err;
 }
 
 
-/* Check the signature of an OCSP repsonse.  OCSP is the context,
+/* Check the signature of an OCSP response.  OCSP is the context,
    S_SIG the signature value and MD the handle of the hash we used for
    the response.  This function automagically finds the correct public
    key.  If SIGNER_FPR_LIST is not NULL, the default OCSP reponder has been
@@ -403,26 +535,11 @@ check_signature (ctrl_t ctrl,
                  fingerprint_list_t signer_fpr_list)
 {
   gpg_error_t err;
-  int algo, cert_idx;
-  gcry_sexp_t s_hash;
+  int cert_idx;
   ksba_cert_t cert;
 
   /* Create a suitable S-expression with the hash value of our response. */
   gcry_md_final (md);
-  algo = gcry_md_get_algo (md);
-  if (algo != GCRY_MD_SHA1 )
-    {
-      log_error (_("only SHA-1 is supported for OCSP responses\n"));
-      return gpg_error (GPG_ERR_DIGEST_ALGO);
-    }
-  err = gcry_sexp_build (&s_hash, NULL, "(data(flags pkcs1)(hash sha1 %b))",
-                         gcry_md_get_algo_dlen (algo),
-                         gcry_md_read (md, algo));
-  if (err)
-    {
-      log_error (_("creating S-expression failed: %s\n"), gcry_strerror (err));
-      return err;
-    }
 
   /* Get rid of old OCSP specific certificate references. */
   release_ctrl_ocsp_certs (ctrl);
@@ -437,13 +554,12 @@ check_signature (ctrl_t ctrl,
         cert = get_cert_local (ctrl, signer_fpr_list->hexfpr);
       if (cert)
         {
-          err = check_signature_core (ctrl, cert, s_sig, s_hash,
+          err = check_signature_core (ctrl, cert, s_sig, md,
                                       signer_fpr_list);
           ksba_cert_release (cert);
           cert = NULL;
           if (!err)
             {
-              gcry_sexp_release (s_hash);
               return 0; /* Successfully verified the signature. */
             }
         }
@@ -461,10 +577,14 @@ check_signature (ctrl_t ctrl,
         {
           cert_ref_t cref;
 
+          /* dump_cert ("from ocsp response", cert); */
           cref = xtrymalloc (sizeof *cref);
           if (!cref)
-            log_error (_("allocating list item failed: %s\n"),
-                       gcry_strerror (err));
+            {
+              err = gpg_error_from_syserror ();
+              log_error (_("allocating list item failed: %s\n"),
+                         gpg_strerror (err));
+            }
           else if (!cache_cert_silent (cert, &cref->fpr))
             {
               cref->next = ctrl->ocsp_certs;
@@ -496,23 +616,32 @@ check_signature (ctrl_t ctrl,
             }
           log_printf ("not found\n");
         }
-      ksba_free (name);
-      ksba_free (keyid);
 
       if (cert)
         {
-          err = check_signature_core (ctrl, cert, s_sig, s_hash,
-                                      signer_fpr_list);
+          err = check_signature_core (ctrl, cert, s_sig, md, signer_fpr_list);
           ksba_cert_release (cert);
           if (!err)
             {
-              gcry_sexp_release (s_hash);
+              ksba_free (name);
+              ksba_free (keyid);
               return 0; /* Successfully verified the signature. */
             }
+          log_error ("responder certificate ");
+          if (name)
+            log_printf ("'/%s' ", name);
+          if (keyid)
+            {
+              log_printf ("{");
+              dump_serial (keyid);
+              log_printf ("} ");
+            }
+          log_printf ("did not verify: %s\n", gpg_strerror (err));
         }
+      ksba_free (name);
+      ksba_free (keyid);
     }
 
-  gcry_sexp_release (s_hash);
   log_error (_("no suitable certificate found to verify the OCSP response\n"));
   return gpg_error (GPG_ERR_NO_PUBKEY);
 }
@@ -521,10 +650,13 @@ check_signature (ctrl_t ctrl,
 /* Check whether the certificate either given by fingerprint CERT_FPR
    or directly through the CERT object is valid by running an OCSP
    transaction.  With FORCE_DEFAULT_RESPONDER set only the configured
-   default responder is used. */
+   default responder is used.  If R_REVOKED_AT or R_REASON are not
+   NULL and the certificat has been revoked the revocation time and
+   the reasons are stored there. */
 gpg_error_t
 ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr,
-              int force_default_responder)
+              int force_default_responder, ksba_isotime_t r_revoked_at,
+              const char **r_reason)
 {
   gpg_error_t err;
   ksba_ocsp_t ocsp = NULL;
@@ -543,6 +675,12 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr,
   char *oid;
   ksba_name_t name;
   fingerprint_list_t default_signer = NULL;
+  const char *sreason;
+
+  if (r_revoked_at)
+    *r_revoked_at = 0;
+  if (r_reason)
+    *r_reason = NULL;
 
   /* Get the certificate.  */
   if (cert)
@@ -584,8 +722,6 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr,
       goto leave;
     }
 
-
-
   /* Figure out the OCSP responder to use.
      1. Try to get the reponder from the certificate.
         We do only take http and https style URIs into account.
@@ -642,20 +778,40 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr,
     }
 
   /* Ask the OCSP responder. */
-  err = gcry_md_open (&md, GCRY_MD_SHA1, 0);
-  if (err)
-    {
-      log_error (_("failed to establish a hashing context for OCSP: %s\n"),
-                 gpg_strerror (err));
-      goto leave;
-    }
-  err = do_ocsp_request (ctrl, ocsp, md, url, cert, issuer_cert);
+  err = do_ocsp_request (ctrl, ocsp, url, cert, issuer_cert,
+                         &sigval, produced_at, &md);
   if (err)
     goto leave;
 
+  /* It is sometimes useful to know the responder ID. */
+  if (opt.verbose)
+    {
+      char *resp_name;
+      ksba_sexp_t resp_keyid;
+
+      err = ksba_ocsp_get_responder_id (ocsp, &resp_name, &resp_keyid);
+      if (err)
+        log_info (_("error getting responder ID: %s\n"), gpg_strerror (err));
+      else
+        {
+          log_info ("responder id: ");
+          if (resp_name)
+            log_printf ("'/%s' ", resp_name);
+          if (resp_keyid)
+            {
+              log_printf ("{");
+              dump_serial (resp_keyid);
+              log_printf ("} ");
+            }
+          log_printf ("\n");
+        }
+      ksba_free (resp_name);
+      ksba_free (resp_keyid);
+      err = 0;
+    }
+
   /* We got a useful answer, check that the answer has a valid signature. */
-  sigval = ksba_ocsp_get_sig_val (ocsp, produced_at);
-  if (!sigval || !*produced_at)
+  if (!sigval || !*produced_at || !md)
     {
       err = gpg_error (GPG_ERR_INV_OBJ);
       goto leave;
@@ -695,8 +851,36 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr,
                       more important message than the failure of our
                       cache. */
         }
-    }
 
+      switch (reason)
+        {
+        case KSBA_CRLREASON_UNSPECIFIED:
+          sreason = "unspecified"; break;
+        case KSBA_CRLREASON_KEY_COMPROMISE:
+          sreason = "key compromise"; break;
+        case KSBA_CRLREASON_CA_COMPROMISE:
+          sreason = "CA compromise"; break;
+        case KSBA_CRLREASON_AFFILIATION_CHANGED:
+          sreason = "affiliation changed"; break;
+        case KSBA_CRLREASON_SUPERSEDED:
+          sreason = "superseded"; break;
+        case KSBA_CRLREASON_CESSATION_OF_OPERATION:
+          sreason = "cessation of operation"; break;
+        case KSBA_CRLREASON_CERTIFICATE_HOLD:
+          sreason = "certificate on hold"; break;
+        case KSBA_CRLREASON_REMOVE_FROM_CRL:
+          sreason = "removed from CRL"; break;
+        case KSBA_CRLREASON_PRIVILEGE_WITHDRAWN:
+          sreason = "privilege withdrawn"; break;
+        case KSBA_CRLREASON_AA_COMPROMISE:
+          sreason = "AA compromise"; break;
+        case KSBA_CRLREASON_OTHER:
+          sreason = "other"; break;
+        default: sreason = "?"; break;
+        }
+    }
+  else
+    sreason = "";
 
   if (opt.verbose)
     {
@@ -708,29 +892,19 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr,
                 this_update, next_update);
       if (status == KSBA_STATUS_REVOKED)
         log_info (_("certificate has been revoked at: %s due to: %s\n"),
-                  revocation_time,
-                  reason == KSBA_CRLREASON_UNSPECIFIED?   "unspecified":
-                  reason == KSBA_CRLREASON_KEY_COMPROMISE? "key compromise":
-                  reason == KSBA_CRLREASON_CA_COMPROMISE?   "CA compromise":
-                  reason == KSBA_CRLREASON_AFFILIATION_CHANGED?
-                                                      "affiliation changed":
-                  reason == KSBA_CRLREASON_SUPERSEDED?   "superseded":
-                  reason == KSBA_CRLREASON_CESSATION_OF_OPERATION?
-                                                  "cessation of operation":
-                  reason == KSBA_CRLREASON_CERTIFICATE_HOLD?
-                                                  "certificate on hold":
-                  reason == KSBA_CRLREASON_REMOVE_FROM_CRL?
-                                                  "removed from CRL":
-                  reason == KSBA_CRLREASON_PRIVILEGE_WITHDRAWN?
-                                                  "privilege withdrawn":
-                  reason == KSBA_CRLREASON_AA_COMPROMISE? "AA compromise":
-                  reason == KSBA_CRLREASON_OTHER?   "other":"?");
+                  revocation_time, sreason);
 
     }
 
 
   if (status == KSBA_STATUS_REVOKED)
-    err = gpg_error (GPG_ERR_CERT_REVOKED);
+    {
+      err = gpg_error (GPG_ERR_CERT_REVOKED);
+      if (r_revoked_at)
+        gnupg_copy_time (r_revoked_at, revocation_time);
+      if (r_reason)
+        *r_reason = sreason;
+    }
   else if (status == KSBA_STATUS_UNKNOWN)
     err = gpg_error (GPG_ERR_NO_DATA);
   else if (status != KSBA_STATUS_GOOD)
@@ -761,7 +935,7 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr,
         err = gpg_error (GPG_ERR_TIME_CONFLICT);
     }
 
-  /* Check that we are not beyound NEXT_UPDATE  (plus some extra time). */
+  /* Check that we are not beyond NEXT_UPDATE  (plus some extra time). */
   if (*next_update)
     {
       gnupg_copy_time (tmp_time, next_update);
