@@ -30,8 +30,10 @@
 #include "options.h"
 
 
-/* This is mpi_copy with a fix for opaque MPIs which store a NULL
-   pointer.  This will also be fixed in Libggcrypt 1.7.0.  */
+/* This is a wrapper for mpi_copy which handles opaque MPIs with a
+ * NULL pointer as opaque data; e.g. gcry_mpi_set_opaque(a, NULL, 0).
+ * It seems that at least gcry_mpi_set_opaque_copy does not yet handle
+ * this correctly.  */
 static gcry_mpi_t
 my_mpi_copy (gcry_mpi_t a)
 {
@@ -77,11 +79,6 @@ free_seckey_enc( PKT_signature *sig )
   xfree(sig->hashed);
   xfree(sig->unhashed);
 
-  if (sig->pka_info)
-    {
-      xfree (sig->pka_info->uri);
-      xfree (sig->pka_info);
-    }
   xfree (sig->signers_uid);
 
   xfree(sig);
@@ -189,10 +186,10 @@ copy_prefs (const prefitem_t *prefs)
 
 
 /* Copy the public key S to D.  If D is NULL allocate a new public key
-   structure.  If S has seckret key infos, only the public stuff is
-   copied.  */
+ * structure.  Only the basic stuff is copied; not any ancillary
+ * data.  */
 PKT_public_key *
-copy_public_key (PKT_public_key *d, PKT_public_key *s)
+copy_public_key_basics (PKT_public_key *d, PKT_public_key *s)
 {
   int n, i;
 
@@ -200,8 +197,8 @@ copy_public_key (PKT_public_key *d, PKT_public_key *s)
     d = xmalloc (sizeof *d);
   memcpy (d, s, sizeof *d);
   d->seckey_info = NULL;
-  d->user_id = scopy_user_id (s->user_id);
-  d->prefs = copy_prefs (s->prefs);
+  d->user_id = NULL;
+  d->prefs = NULL;
 
   n = pubkey_get_npkey (s->pubkey_algo);
   i = 0;
@@ -215,6 +212,24 @@ copy_public_key (PKT_public_key *d, PKT_public_key *s)
   for (; i < PUBKEY_MAX_NSKEY; i++)
     d->pkey[i] = NULL;
 
+  d->revkey = NULL;
+  d->serialno = NULL;
+  d->updateurl = NULL;
+
+  return d;
+}
+
+
+/* Copy the public key S to D.  If D is NULL allocate a new public key
+   structure.  If S has seckret key infos, only the public stuff is
+   copied.  */
+PKT_public_key *
+copy_public_key (PKT_public_key *d, PKT_public_key *s)
+{
+  d = copy_public_key_basics (d, s);
+  d->user_id = scopy_user_id (s->user_id);
+  d->prefs = copy_prefs (s->prefs);
+
   if (!s->revkey && s->numrevkeys)
     BUG();
   if (s->numrevkeys)
@@ -222,8 +237,6 @@ copy_public_key (PKT_public_key *d, PKT_public_key *s)
       d->revkey = xmalloc(sizeof(struct revocation_key)*s->numrevkeys);
       memcpy(d->revkey,s->revkey,sizeof(struct revocation_key)*s->numrevkeys);
     }
-  else
-    d->revkey = NULL;
 
   if (s->serialno)
     d->serialno = xstrdup (s->serialno);
@@ -235,20 +248,6 @@ copy_public_key (PKT_public_key *d, PKT_public_key *s)
 
 
 
-static pka_info_t *
-cp_pka_info (const pka_info_t *s)
-{
-  pka_info_t *d = xmalloc (sizeof *s + strlen (s->email));
-
-  d->valid = s->valid;
-  d->checked = s->checked;
-  d->uri = s->uri? xstrdup (s->uri):NULL;
-  memcpy (d->fpr, s->fpr, sizeof s->fpr);
-  strcpy (d->email, s->email);
-  return d;
-}
-
-
 PKT_signature *
 copy_signature( PKT_signature *d, PKT_signature *s )
 {
@@ -264,7 +263,6 @@ copy_signature( PKT_signature *d, PKT_signature *s )
 	for(i=0; i < n; i++ )
 	    d->data[i] = my_mpi_copy( s->data[i] );
     }
-    d->pka_info = s->pka_info? cp_pka_info (s->pka_info) : NULL;
     d->hashed = cp_subpktarea (s->hashed);
     d->unhashed = cp_subpktarea (s->unhashed);
     if (s->signers_uid)
@@ -472,6 +470,7 @@ free_packet (PACKET *pkt, parse_packet_ctx_t parsectx)
       break;
     case PKT_ENCRYPTED:
     case PKT_ENCRYPTED_MDC:
+    case PKT_ENCRYPTED_AEAD:
       free_encrypted (pkt->pkt.encrypted);
       break;
     case PKT_PLAINTEXT:
