@@ -1,6 +1,6 @@
 /* server.c - Server mode and main entry point
- * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
- *               2010 Free Software Foundation, Inc.
+ * Copyright (C) 2001-2010 Free Software Foundation, Inc.
+ * Copyright (C) 2001-2011, 2013-2020 g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -31,6 +31,8 @@
 #include <assuan.h>
 #include "../common/sysutils.h"
 #include "../common/server-help.h"
+#include "../common/asshelp.h"
+#include "../common/shareddefs.h"
 
 #define set_error(e,t) assuan_set_error (ctx, gpg_error (e), (t))
 
@@ -132,9 +134,6 @@ close_message_fd (ctrl_t ctrl)
 {
   if (ctrl->server_local->message_fd != -1)
     {
-#ifdef HAVE_W32CE_SYSTEM
-#warning Is this correct for W32/W32CE?
-#endif
       close (ctrl->server_local->message_fd);
       ctrl->server_local->message_fd = -1;
     }
@@ -288,6 +287,32 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
           ctrl->offline = i;
         }
     }
+  else if (!strcmp (key, "always-trust"))
+    {
+      /* We ignore this option if gpgsm has been started with
+         --always-trust (which also sets offline) and if
+         --require-compliance is active */
+      if (!opt.always_trust && !opt.require_compliance)
+        {
+          int i = *value? !!atoi (value) : 1;
+          ctrl->always_trust = i;
+        }
+    }
+  else if (!strcmp (key, "request-origin"))
+    {
+      if (!opt.request_origin)
+        {
+          int i = parse_request_origin (value);
+          if (i == -1)
+            err = gpg_error (GPG_ERR_INV_VALUE);
+          else
+            opt.request_origin = i;
+        }
+    }
+  else if (!strcmp (key, "input-size-hint"))
+    {
+      ctrl->input_size_hint = string_to_u64 (value);
+    }
   else
     err = gpg_error (GPG_ERR_UNKNOWN_OPTION);
 
@@ -306,6 +331,7 @@ reset_notify (assuan_context_t ctx, char *line)
   gpgsm_release_certlist (ctrl->server_local->signerlist);
   ctrl->server_local->recplist = NULL;
   ctrl->server_local->signerlist = NULL;
+  ctrl->always_trust = 0;
   close_message_fd (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
@@ -474,6 +500,7 @@ cmd_encrypt (assuan_context_t ctx, char *line)
 
   gpgsm_release_certlist (ctrl->server_local->recplist);
   ctrl->server_local->recplist = NULL;
+  ctrl->always_trust = 0;
   /* Close and reset the fd */
   close_message_fd (ctrl);
   assuan_close_input_fd (ctx);
@@ -711,8 +738,13 @@ cmd_export (assuan_context_t ctx, char *line)
 
   if (opt_secret)
     {
-      if (!list || !*list->d)
+      if (!list)
         return set_error (GPG_ERR_NO_DATA, "No key given");
+      if (!*list->d)
+        {
+          free_strlist (list);
+          return set_error (GPG_ERR_NO_DATA, "No key given");
+        }
       if (list->next)
         return set_error (GPG_ERR_TOO_MANY, "Only one key allowed");
   }
@@ -853,12 +885,6 @@ cmd_message (assuan_context_t ctx, char *line)
   if (rc)
     return rc;
 
-#ifdef HAVE_W32CE_SYSTEM
-  sysfd = _assuan_w32ce_finish_pipe ((int)sysfd, 0);
-  if (sysfd == INVALID_HANDLE_VALUE)
-    return set_error (gpg_err_code_from_syserror (),
-		      "rvid conversion failed");
-#endif
 
   fd = translate_sys2libc_fd (sysfd, 0);
   if (fd == -1)
@@ -870,10 +896,10 @@ cmd_message (assuan_context_t ctx, char *line)
 
 
 static const char hlp_listkeys[] =
-  "LISTKEYS [<patterns>]\n"
-  "LISTSECRETKEYS [<patterns>]\n"
-  "DUMPKEYS [<patterns>]\n"
-  "DUMPSECRETKEYS [<patterns>]\n"
+  "LISTKEYS       [<options>] [<patterns>]\n"
+  "LISTSECRETKEYS [<options>] [<patterns>]\n"
+  "DUMPKEYS       [<options>] [<patterns>]\n"
+  "DUMPSECRETKEYS [<options>] [<patterns>]\n"
   "\n"
   "List all certificates or only those specified by PATTERNS.  Each\n"
   "pattern shall be a percent-plus escaped certificate specification.\n"
@@ -882,8 +908,12 @@ static const char hlp_listkeys[] =
   "smartcard has been registered.  The \"DUMP\" versions of the command\n"
   "are only useful for debugging.  The output format is a percent escaped\n"
   "colon delimited listing as described in the manual.\n"
+  "Supported values for OPTIONS are:\n"
+  "  --           Stop option processing\n"
+  "  --issuer-der PATTERN is a DER of the serialnumber as hexstring;\n"
+  "               the issuer is then inquired with \"ISSUER_DER\".\n"
   "\n"
-  "These \"OPTION\" command keys effect the output::\n"
+  "These Assuan \"OPTION\" command keys effect the output::\n"
   "\n"
   "  \"list-mode\" set to 0: List only local certificates (default).\n"
   "                     1: Ditto.\n"
@@ -903,9 +933,14 @@ do_listkeys (assuan_context_t ctx, char *line, int mode)
   ctrl_t ctrl = assuan_get_pointer (ctx);
   estream_t fp;
   char *p;
+  size_t n;
   strlist_t list, sl;
   unsigned int listmode;
   gpg_error_t err;
+  int opt_issuer_der;
+
+  opt_issuer_der = has_option (line, "--issuer-der");
+  line = skip_options (line);
 
   /* Break the line down into an strlist. */
   list = NULL;
@@ -929,23 +964,90 @@ do_listkeys (assuan_context_t ctx, char *line, int mode)
           list = sl;
         }
     }
+  if (opt_issuer_der && (!list || list->next))
+    {
+      free_strlist (list);
+      return set_error (GPG_ERR_INV_ARG,
+                        "only one arg for --issuer-der please");
+    }
+
+  if (opt_issuer_der)
+    {
+      unsigned char *value = NULL;
+      size_t valuelen;
+      char *issuer;
+
+      err = assuan_inquire (ctx, "ISSUER_DER", &value, &valuelen, 0);
+      if (err)
+        {
+          free_strlist (list);
+          return err;
+        }
+      if (!valuelen)
+        {
+          xfree (value);
+          free_strlist (list);
+          return gpg_error (GPG_ERR_MISSING_VALUE);
+        }
+      err = ksba_dn_der2str (value, valuelen, &issuer);
+      xfree (value);
+      if (err)
+        {
+          free_strlist (list);
+          return err;
+        }
+      /* ksba_dn_der2str seems to always append "\\0A".  Trim that.  */
+      n = strlen (issuer);
+      if (n > 3 && !strcmp (issuer + n - 3, "\\0A"))
+        issuer[n-3] = 0;
+
+      p = strconcat ("#", list->d, "/", issuer, NULL);
+      if (!p)
+        {
+          err = gpg_error_from_syserror ();
+          ksba_free (issuer);
+          free_strlist (list);
+          return err;
+        }
+      ksba_free (issuer);
+      free_strlist (list);
+      list = NULL;
+      if (!add_to_strlist_try (&list, p))
+        {
+          err = gpg_error_from_syserror ();
+          xfree (p);
+          return err;
+        }
+      xfree (p);
+    }
+
 
   if (ctrl->server_local->list_to_output)
     {
       int outfd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
 
       if ( outfd == -1 )
-        return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
+        {
+          free_strlist (list);
+          return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
+        }
       fp = es_fdopen_nc (outfd, "w");
       if (!fp)
-        return set_error (gpg_err_code_from_syserror (), "es_fdopen() failed");
+        {
+          free_strlist (list);
+          return set_error (gpg_err_code_from_syserror (),
+                            "es_fdopen() failed");
+        }
     }
   else
     {
       fp = es_fopencookie (ctx, "w", data_line_cookie_functions);
       if (!fp)
-        return set_error (GPG_ERR_ASS_GENERAL,
-                          "error setting up a data stream");
+        {
+          free_strlist (list);
+          return set_error (GPG_ERR_ASS_GENERAL,
+                            "error setting up a data stream");
+        }
     }
 
   ctrl->with_colons = 1;
@@ -955,6 +1057,7 @@ do_listkeys (assuan_context_t ctx, char *line, int mode)
   if (ctrl->server_local->list_external)
     listmode |= (1<<7);
   err = gpgsm_list_keys (assuan_get_pointer (ctx), list, fp, listmode);
+
   free_strlist (list);
   es_fclose (fp);
   if (ctrl->server_local->list_to_output)
@@ -1099,7 +1202,8 @@ static const char hlp_getinfo[] =
   "  agent-check - Return success if the agent is running.\n"
   "  cmd_has_option CMD OPT\n"
   "              - Returns OK if the command CMD implements the option OPT.\n"
-  "  offline     - Returns OK if the connection is in offline mode.";
+  "  offline     - Returns OK if the connection is in offline mode."
+  "  always-trust- Returns OK if the connection is in always-trust mode.";
 static gpg_error_t
 cmd_getinfo (assuan_context_t ctx, char *line)
 {
@@ -1149,14 +1253,19 @@ cmd_getinfo (assuan_context_t ctx, char *line)
                 {
                   cmdopt = line;
                   if (!command_has_option (cmd, cmdopt))
-                    rc = gpg_error (GPG_ERR_GENERAL);
+                    rc = gpg_error (GPG_ERR_FALSE);
                 }
             }
         }
     }
   else if (!strcmp (line, "offline"))
     {
-      rc = ctrl->offline? 0 : gpg_error (GPG_ERR_GENERAL);
+      rc = ctrl->offline? 0 : gpg_error (GPG_ERR_FALSE);
+    }
+  else if (!strcmp (line, "always-trust"))
+    {
+      rc = (ctrl->always_trust || opt.always_trust)? 0
+           /**/                                    : gpg_error (GPG_ERR_FALSE);
     }
   else
     rc = set_error (GPG_ERR_ASS_PARAMETER, "unknown value for WHAT");
@@ -1275,13 +1384,9 @@ gpgsm_server (certlist_t default_recplist)
   /* We use a pipe based server so that we can work from scripts.
      assuan_init_pipe_server will automagically detect when we are
      called with a socketpair and ignore FILEDES in this case. */
-#ifdef HAVE_W32CE_SYSTEM
-  #define SERVER_STDIN es_fileno(es_stdin)
-  #define SERVER_STDOUT es_fileno(es_stdout)
-#else
 #define SERVER_STDIN 0
 #define SERVER_STDOUT 1
-#endif
+
   filedes[0] = assuan_fdopen (SERVER_STDIN);
   filedes[1] = assuan_fdopen (SERVER_STDOUT);
   rc = assuan_new (&ctx);
@@ -1371,6 +1476,8 @@ gpgsm_server (certlist_t default_recplist)
   audit_release (ctrl.audit);
   ctrl.audit = NULL;
 
+  gpgsm_deinit_default_ctrl (&ctrl);
+
   assuan_release (ctx);
 }
 
@@ -1422,28 +1529,19 @@ gpgsm_status2 (ctrl_t ctrl, int no, ...)
             }
         }
       putc ('\n', statusfp);
-      fflush (statusfp);
+      if (ferror (statusfp))
+        err = gpg_error_from_syserror ();
+      else
+        {
+          fflush (statusfp);
+          if (ferror (statusfp))
+            err = gpg_error_from_syserror ();
+        }
     }
   else
     {
-      assuan_context_t ctx = ctrl->server_local->assuan_ctx;
-      char buf[950], *p;
-      size_t n;
-
-      p = buf;
-      n = 0;
-      while ( (text = va_arg (arg_ptr, const char *)) )
-        {
-          if (n)
-            {
-              *p++ = ' ';
-              n++;
-            }
-          for ( ; *text && n < DIM (buf)-2; n++)
-            *p++ = *text++;
-        }
-      *p = 0;
-      err = assuan_write_status (ctx, get_status_string (no), buf);
+      err = vprint_assuan_status_strings (ctrl->server_local->assuan_ctx,
+                                          get_status_string (no), arg_ptr);
     }
 
   va_end (arg_ptr);
@@ -1462,7 +1560,7 @@ gpgsm_status_with_err_code (ctrl_t ctrl, int no, const char *text,
 {
   char buf[30];
 
-  sprintf (buf, "%u", (unsigned int)ec);
+  snprintf (buf, sizeof buf, "%u", (unsigned int)ec);
   if (text)
     return gpgsm_status2 (ctrl, no, text, buf, NULL);
   else
@@ -1483,6 +1581,45 @@ gpgsm_status_with_error (ctrl_t ctrl, int no, const char *text,
 }
 
 
+/* This callback is used to emit progress status lines.  */
+gpg_error_t
+gpgsm_progress_cb (ctrl_t ctrl, uint64_t current, uint64_t total)
+{
+  char buffer[60];
+  char units[] = "BKMGTPEZY?";
+  int unitidx = 0;
+
+  if (total)
+    {
+      if (current > total)
+        current = total;
+
+      while (total > 1024*1024)
+        {
+          total /= 1024;
+          current /= 1024;
+          unitidx++;
+        }
+    }
+  else
+    {
+      while (current > 1024*1024)
+        {
+          current /= 1024;
+          unitidx++;
+        }
+    }
+
+  if (unitidx > 9)
+    unitidx = 9;
+
+  snprintf (buffer, sizeof buffer, "? %lu %lu %c%s",
+            (unsigned long)current, (unsigned long)total,
+            units[unitidx], unitidx? "iB" : "");
+  return gpgsm_status2 (ctrl, STATUS_PROGRESS, "?", buffer, NULL);
+}
+
+
 /* Helper to notify the client about Pinentry events.  Because that
    might disturb some older clients, this is only done when enabled
    via an option.  Returns an gpg error code. */
