@@ -33,7 +33,10 @@
 
 (define alpha "Alpha <alpha@invalid.example.net>")
 (define bravo "Bravo <bravo@invalid.example.net>")
+(define bravohash "F75B1420CC5881F4005333379355CFF5873094DE")
 (define charlie "Charlie <charlie@invalid.example.net>")
+(define delta "Delta <delta@invalid.example.net>")
+(define deltahash "359DC5EFF98B14A58AAA615C638E8BD0CEDA537B")
 
 (define (key-data key)
   (filter (lambda (x) (or (string=? (car x) "pub")
@@ -75,7 +78,8 @@
 
 (info "Checking that we can mark an user ID as primary.")
 (call-check `(,@gpg --quick-set-primary-uid ,(exact alpha) ,alpha))
-(call-check `(,@gpg --quick-set-primary-uid ,(exact alpha) ,bravo))
+(info "Checking that we can mark a user ID as primary by its hash...")
+(call-check `(,@gpg --quick-set-primary-uid ,(exact alpha) ,bravohash))
 ;; XXX I don't know how to verify this.  The keylisting does not seem
 ;; to indicate the primary UID.
 
@@ -87,6 +91,10 @@
 (info "Checking that we can revoke a user ID...")
 (call-check `(,@GPG --quick-revoke-uid ,(exact bravo) ,alpha))
 
+(info "Checking that we can revoke a user ID by its hash...")
+(call-check `(,@GPG --quick-add-uid ,(exact bravo) ,delta))
+(call-check `(,@GPG --quick-revoke-uid ,(exact bravo) ,deltahash))
+
 (info "Checking that we get an error revoking a non-existent user ID.")
 (catch '()
        (call-check `(,@GPG --quick-revoke-uid ,(exact bravo) ,charlie))
